import {
  ConflictException,
  ForbiddenException,
  forwardRef,
  Inject,
  Injectable,
  NotFoundException,
} from '@nestjs/common';
import { InjectRepository } from '@nestjs/typeorm';
import { FindOptionsWhere, Not, Repository } from 'typeorm';
import { Carrito } from '../carrito/entity/carrito.entity';
import { Equipo } from './entity/equipo.entity';
import { Institucion } from '../institucion/entity/institucion.entity';
import { Marca } from '../marca/entity/marca.entity';
import { Modelo } from '../modelo/entity/modelo.entity';
import { Modulo } from '../modulo/entity/modulo.entity';
import { Operador } from '../operador/entity/operador.entity';
import { Programa } from '../institucion-programa/entity/programa.entity';
import { TipoCarrito } from '../institucion-tipo-carrito/entity/tipo-carrito.entity';
import { TipoEntrada } from '../institucion-tipo-entrada/entity/tipo-entrada.entity';
import { Usuario } from '../usuario/entity/usuario.entity';
import { FullInformacionEquipoView } from './entity/views/full-informacion-equipo.view';
import { InformacionEquipoView } from './entity/views/informacion-equipo.view';
import { InformacionEquipoPrestamoView } from './entity/views/informacion-equipo-prestamo.view';
import { CarritoService } from '../carrito/carrito.service';
import { EquipoMotivoService } from '../equipo-motivo/equipo-motivo.service';
import { EquipoProgramaService } from '../equipo-programa/equipo-programa.service';
import { EquipoTipoEntradaService } from '../equipo-tipo-entrada/equipo-tipo-entrada.service';
import { InstitucionService } from '../institucion/institucion.service';
import { InstitucionProgramaService } from '../institucion-programa/institucion-programa.service';
import { InstitucionTipoCarritoService } from '../institucion-tipo-carrito/institucion-tipo-carrito.service';
import { InstitucionTipoEntradaService } from '../institucion-tipo-entrada/institucion-tipo-entrada.service';
import { MarcaService } from '../marca/marca.service';
import { ModeloService } from '../modelo/modelo.service';
import { ModuloService } from '../modulo/modulo.service';
import { StatusService } from '../status/status.service';

@Injectable()
export class EquipoService {
  constructor(
    @InjectRepository(Equipo) private repository: Repository<Equipo>,
    @InjectRepository(FullInformacionEquipoView)
    private fullInformacionEquipoView: Repository<FullInformacionEquipoView>,
    @InjectRepository(InformacionEquipoView)
    private informacionEquipoView: Repository<InformacionEquipoView>,
    @InjectRepository(InformacionEquipoPrestamoView)
    private informacionEquipoPrestamoView: Repository<InformacionEquipoPrestamoView>,
    private carritoService: CarritoService,
    @Inject(forwardRef(() => EquipoMotivoService))
    private equipoMotivoService: EquipoMotivoService,
    @Inject(forwardRef(() => EquipoProgramaService))
    private equipoProgramaService: EquipoProgramaService,
    @Inject(forwardRef(() => EquipoTipoEntradaService))
    private equipoTipoEntradaService: EquipoTipoEntradaService,
    private institucionService: InstitucionService,
    private institucionProgramaService: InstitucionProgramaService,
    private institucionTipoCarritoService: InstitucionTipoCarritoService,
    private institucionTipoEntradaService: InstitucionTipoEntradaService,
    private marcaService: MarcaService,
    private modeloService: ModeloService,
    private moduloService: ModuloService,
    private statusService: StatusService,
  ) {}

  async create(
    equipo: string,
    carrito: Carrito,
    numero_inventario: string,
    numero_serie: string,
    marca: Marca,
    modelo: Modelo,
  ): Promise<Equipo> {
    // Creamos y guardamos un registro
    return this.repository
      .save(
        this.repository.create({
          equipo,
          numero_inventario,
          numero_serie,
          carrito,
          marca,
          modelo,
        }),
      )
      .then(async (equipo) => {
        // Asignamos programa "Sin programa" por default
        await this.equipoProgramaService.createSinPrograma(equipo);
        return equipo;
      });
  }

  async findAll(
    operador: Operador,
    filtros: {
      pagina: string;
      carrito?: string;
      equipo?: string;
      id_carrito?: string;
      id_institucion?: string;
      id_marca?: string;
      id_modelo?: string;
      id_modulo?: string;
      id_programa?: string;
      id_status?: string;
      id_tipo_carrito?: string;
      id_tipo_entrada?: string;
    },
  ): Promise<[Equipo[], number]> {
    const carrito = filtros.id_carrito
      ? await this.carritoService.findById(parseInt(filtros.id_carrito))
      : null;
    const institucion = filtros.id_institucion
      ? await this.institucionService.findById(parseInt(filtros.id_institucion))
      : null;
    const marca = filtros.id_marca
      ? await this.marcaService.findById(parseInt(filtros.id_marca), 'e')
      : null;
    const modelo = filtros.id_modelo
      ? await this.modeloService.findById(parseInt(filtros.id_modelo), 'e')
      : null;
    const modulo = filtros.id_modulo
      ? await this.moduloService.findById(parseInt(filtros.id_modulo))
      : null;
    const programa = filtros.id_programa
      ? await this.institucionProgramaService.findProgramaById(
          parseInt(filtros.id_programa),
        )
      : null;
    const tipoCarrito = filtros.id_tipo_carrito
      ? await this.institucionTipoCarritoService.findTipoCarritoById(
          parseInt(filtros.id_tipo_carrito),
        )
      : null;
    const tipoEntrada = filtros.id_tipo_entrada
      ? await this.institucionTipoEntradaService.findTipoEntradaById(
          parseInt(filtros.id_tipo_entrada),
        )
      : null;
    const status = filtros.id_status
      ? await this.statusService.findById(parseInt(filtros.id_status))
      : null;
    const query = this.repository
      .createQueryBuilder('e')
      .innerJoinAndSelect('e.carrito', 'c')
      .innerJoinAndSelect('e.marca', 'ma')
      .innerJoinAndSelect('e.modelo', 'mo')
      .innerJoinAndSelect('e.programas', 'ps')
      .innerJoinAndSelect('e.status', 's')
      .innerJoinAndSelect('e.tiposEntradas', 'tes')
      .innerJoinAndSelect('c.modulo', 'm')
      .innerJoinAndSelect('c.tipoCarrito', 'tc')
      .innerJoinAndSelect('ps.programa', 'p')
      .innerJoinAndSelect('tes.tipoEntrada', 'te')
      .innerJoinAndSelect('m.institucion', 'i')
      .orderBy('i.institucion')
      .addOrderBy('m.modulo')
      .addOrderBy('tc.tipo_carrito')
      .addOrderBy('c.carrito')
      .addOrderBy('e.equipo')
      .take(15)
      .skip((parseInt(filtros.pagina) - 1) * 15);

    if (operador.tipoUsuario.id_tipo_usuario > 2)
      query.andWhere('m.id_institucion = :id_institucion', {
        id_institucion: operador.institucion.id_institucion,
      });
    else if (institucion)
      query.andWhere('m.id_institucion = :id_institucion', {
        id_institucion: institucion.id_institucion,
      });
    if (modulo)
      query.andWhere('m.id_modulo = :id_modulo', {
        id_modulo: modulo.id_modulo,
      });
    if (carrito)
      query.andWhere('c.id_carrito = :id_carrito', {
        id_carrito: carrito.id_carrito,
      });
    if (filtros.carrito)
      query.andWhere('c.carrito LIKE :carrito', {
        carrito: `%${filtros.carrito}%`,
      });
    if (filtros.equipo)
      query.andWhere('e.equipo LIKE :equipo', {
        equipo: `%${filtros.equipo}%`,
      });
    if (marca)
      query.andWhere('ma.id_marca = :id_marca', {
        id_marca: marca.id_marca,
      });
    if (modelo)
      query.andWhere('mo.id_modelo = :id_modelo', {
        id_modelo: modelo.id_modelo,
      });
    if (programa)
      query.andWhere('ps.id_programa = :id_programa', {
        id_programa: programa.id_programa,
      });
    if (tipoCarrito)
      query.andWhere('tc.id_tipo_carrito = :id_tipo_carrito', {
        id_tipo_carrito: tipoCarrito.id_tipo_carrito,
      });
    if (tipoEntrada)
      query.andWhere('tes.id_tipo_entrada = :id_tipo_entrada', {
        id_tipo_entrada: tipoEntrada.id_tipo_entrada,
      });
    if (status)
      query.andWhere('s.id_status = :id_status', {
        id_status: status.id_status,
      });
    return query.getManyAndCount();
  }

  findAllResetear(): Promise<Equipo[]> {
    return this.informacionEquipoView
      .find({ where: { prestado: 1, id_status: Not(3) } })
      .then((infoEquipos) => {
        const equipos: Equipo[] = [];

        for (let i = 0; i < infoEquipos.length; i++)
          equipos.push(this.viewToEquipo(infoEquipos[i]));
        return equipos;
      });
  }

  findByEquipo(
    carrito: Carrito,
    equipo: string,
    validarNoExiste = true,
  ): Promise<Equipo> {
    return this.informacionEquipoView
      .findOne({ where: { id_carrito: carrito.id_carrito, equipo } })
      .then((infoEquipo) => {
        if (!infoEquipo && validarNoExiste)
          throw new NotFoundException(
            'No existe un equipo de cómputo con este número de inventario.',
          );
        return this.viewToEquipo(infoEquipo);
      });
  }

  findById(id_equipo: number): Promise<Equipo> {
    return this.informacionEquipoView
      .findOne({ where: { id_equipo } })
      .then((infoEquipo) => {
        if (!infoEquipo)
          throw new NotFoundException('No existe este id equipo.');
        return this.viewToEquipo(infoEquipo);
      });
  }

  findByNumeroInventario(
    institucion: Institucion,
    numero_inventario: string,
    validarNoExiste = true,
  ): Promise<Equipo> {
    return this.informacionEquipoView
      .findOne({
        where: {
          id_institucion: institucion.id_institucion,
          numero_inventario,
        },
      })
      .then((infoEquipo) => {
        if (!infoEquipo && validarNoExiste)
          throw new NotFoundException(
            'No existe un equipo de cómputo con este número de inventario en esta institución.',
          );
        return this.viewToEquipo(infoEquipo);
      });
  }

  findEquipo(
    usuario: Usuario,
    modulo: Modulo,
    tipoCarrito: TipoCarrito,
    programa?: Programa,
    tipoEntrada?: TipoEntrada,
  ): Promise<Equipo> {
    const busquedaReseteo: FindOptionsWhere<InformacionEquipoPrestamoView> = {
      prestado: 1,
      id_modulo: modulo.id_modulo,
      id_programa: 1,
      id_tipo_carrito: tipoCarrito.id_tipo_carrito,
    };
    const busqueda: FindOptionsWhere<InformacionEquipoPrestamoView> = {
      prestado: 0,
      id_modulo: modulo.id_modulo,
      id_programa: 1,
      id_tipo_carrito: tipoCarrito.id_tipo_carrito,
    };

    // Si se mando un programa se añade a la busqueda
    if (programa) {
      busqueda.id_programa = programa.id_programa;
      busquedaReseteo.id_programa = programa.id_programa;
    }
    // Si se mando un tipo de entrada se añade a la busqueda
    if (tipoEntrada) {
      busqueda.id_tipo_entrada = tipoEntrada.id_tipo_entrada;
      busquedaReseteo.id_tipo_entrada = tipoEntrada.id_tipo_entrada;
    }
    // Intentamos encontrar un equipo
    return this.informacionEquipoPrestamoView
      .findOne({ where: busqueda })
      .then(async (infoEquipo) => {
        // Si no se encuentra un equipo reseteamos todos los equipos con las carecteríasticas solicitadas
        if (!infoEquipo)
          await this.informacionEquipoPrestamoView
            .find({ where: busquedaReseteo })
            .then(async (infoEquipos) => {
              // Reseteamos prestado y guardamos los cambios
              for (let i = 0; i < infoEquipos.length; i++)
                await this.repository.save(
                  this.repository.create({
                    id_equipo: infoEquipos[i].id_equipo,
                    prestado: false,
                  }),
                );
              // Intentamos buscar de nuevo
              return this.informacionEquipoPrestamoView.findOne({
                where: busqueda,
              });
            })
            .then((infoEquipoAux) => {
              // Si no hay saca error
              if (!infoEquipoAux)
                throw new ConflictException(
                  'No hay un equipo de cómputo que cumpla con las caracteríasticas solicitadas o ya no hay equipos disponibles en este momento. Intenta más tarde o cambia las características.',
                );
              infoEquipo = infoEquipoAux;
            });
        // Esta variable debe estar siempre en null antes de asignar el equipo
        if (infoEquipo.u)
          throw new ConflictException(
            'Ocurrio un error al tratar de asignarte un equipo de cómputo, intenta de nuevo.',
          );
        // Cambiamos el equipo a apartado y guardamos el id del usuario al que le asignamos ese equipo
        return this.repository.save(
          this.repository.create({
            id_equipo: infoEquipo.id_equipo,
            u: usuario.id_usuario,
            status: { id_status: 2 },
          }),
        );
      })
      .then((equipo) => this.findFullInfoById(equipo.id_equipo)) // Una vez actualizado lo busco de nuevo
      .then((equipo) => {
        // Si el campo u no coincide con el id_usaurio saca error
        if (equipo.u != usuario.id_usuario)
          throw new ConflictException(
            'Ocurrió un error al tratar de asignarte un equipo de cómputo, intenta de nuevo.',
          );
        return equipo;
      });
  }

  findFullInfoById(id_equipo: number): Promise<Equipo> {
    return this.fullInformacionEquipoView
      .findOne({ where: { id_equipo } })
      .then((infoEquipo) => {
        if (!infoEquipo)
          throw new NotFoundException(
            'No existe un equipo de cómputo con este número de inventario en esta institución.',
          );
        return this.fullViewToEquipo(infoEquipo);
      });
  }

  findFullInfoByNumeroInventario(
    operador: Operador,
    numero_inventario: string,
  ): Promise<Equipo> {
    return this.fullInformacionEquipoView
      .findOne({
        where: {
          id_institucion: operador.institucion.id_institucion,
          numero_inventario,
        },
      })
      .then(async (infoEquipo) => {
        if (!infoEquipo)
          throw new NotFoundException(
            'No existe un equipo de cómputo con este número de inventario en esta institución.',
          );

        const equipo = this.fullViewToEquipo(infoEquipo);
        // Buscamos todos los programas de este equipo
        const programas =
          await this.equipoProgramaService.findFullInfoAllByIdEquipo(
            equipo.id_equipo,
          );
        // Buscamos todos los tipos de entrada de este equipo
        const tiposEntradas =
          await this.equipoTipoEntradaService.findFullInfoAllByIdEquipo(
            equipo.id_equipo,
          );

        equipo.programas = programas;
        equipo.tiposEntradas = tiposEntradas;
        return equipo;
      });
  }

  private fullViewToEquipo(infoEquipo: FullInformacionEquipoView): Equipo {
    return this.repository.create({
      id_equipo: infoEquipo.id_equipo,
      equipo: infoEquipo.equipo,
      numero_inventario: infoEquipo.numero_inventario,
      numero_serie: infoEquipo.numero_serie,
      prestado: infoEquipo.prestado,
      u: infoEquipo.u,
      carrito: {
        id_carrito: infoEquipo.id_carrito,
        carrito: infoEquipo.carrito,
        modulo: {
          id_modulo: infoEquipo.id_modulo,
          modulo: infoEquipo.modulo,
          institucion: {
            id_institucion: infoEquipo.id_institucion,
            institucion: infoEquipo.institucion,
          },
        },
        tipoCarrito: {
          id_tipo_carrito: infoEquipo.id_tipo_carrito,
          tipo_carrito: infoEquipo.tipo_carrito,
        },
      },
      marca: {
        id_marca: infoEquipo.id_marca,
        marca: infoEquipo.marca,
      },
      modelo: {
        id_modelo: infoEquipo.id_modelo,
        modelo: infoEquipo.modelo,
      },
      status: {
        id_status: infoEquipo.id_status,
        status: infoEquipo.status,
      },
    });
  }

  save(equipo: Equipo): Promise<Equipo> {
    return this.repository.save(equipo);
  }

  update(
    operador: Operador,
    attrs: Partial<Equipo>,
    id_carrito?: number,
    id_status?: number,
    motivo?: string,
  ): Promise<{ message: string }> {
    return this.findById(attrs.id_equipo)
      .then(async (equipo) => {
        const carrito = id_carrito
          ? await this.carritoService.findById(id_carrito)
          : null;
        const status = id_status
          ? await this.statusService.findById(id_status)
          : null;

        // Validamos que el equipo pertenezca a la institución del operador
        if (
          operador.institucion.id_institucion !=
          equipo.carrito.modulo.institucion.id_institucion
        )
          throw new ForbiddenException(
            'No puedes modificar la información este equipo porque no pertenece a tu institución.',
          );
        // Si se mandó id_status
        if (status) {
          // No se puede actualziar el status del equipo cuando esta apartado o en uso
          if (equipo.status.id_status === 2 || equipo.status.id_status === 3)
            throw new ConflictException(
              'No se puede cambiar el status de un equipo si esta apartada o en uso.',
            );
          // El operador no puede cambiar el status de un equipo a apartado o en uso
          if (status.id_status === 2 || status.id_status === 3)
            throw new ConflictException(
              'No se puede cambiar manualmente el status de un equipo a apartado o en uso.',
            );
          // Si no se mando motivo saca error
          if (!motivo)
            throw new ConflictException(
              'No se mandó el motivo para el cambio de status.',
            );
        }
        // Asignamos valores enviados al objeto
        Object.assign(equipo, attrs);
        if (status) {
          equipo.status = status;
          // Se crea un registro del cambio de status de este equipo
          await this.equipoMotivoService.create(
            equipo,
            operador,
            status,
            motivo,
          );
        }
        if (carrito) equipo.carrito = carrito;
        // Guardamos los cambios
        return this.repository.save(equipo);
      })
      .then((_) => ({ message: 'Se guardaron los cambios correctamente.' }));
  }

  async updateStatus(
    equipo: Equipo,
    operador?: Operador,
    motivo?: string,
  ): Promise<Equipo> {
    // Si el status al que se esta actualizando es mayor a 3
    if (equipo.status.id_status > 3) {
      // Si no se manda operador saca error
      if (!operador) throw new ConflictException('No se mandó el id operador.');
      // Si no se mando motivo saca error
      if (!motivo)
        throw new ConflictException(
          'No se mandó el motivo de cambio de status.',
        );
      // Se crea un registro del cambio de status de este equipo
      await this.equipoMotivoService.create(
        equipo,
        operador,
        equipo.status,
        motivo,
      );
    }
    // Guardamos los cambios
    return this.repository.save(equipo);
  }

  private viewToEquipo(infoEquipo: InformacionEquipoView): Equipo {
    if (!infoEquipo) return null;
    return this.repository.create({
      id_equipo: infoEquipo.id_equipo,
      equipo: infoEquipo.equipo,
      numero_inventario: infoEquipo.numero_inventario,
      prestado: infoEquipo.prestado === 1,
      u: infoEquipo.u,
      carrito: {
        id_carrito: infoEquipo.id_carrito,
        carrito: infoEquipo.carrito,
        modulo: {
          id_modulo: infoEquipo.id_modulo,
          institucion: { id_institucion: infoEquipo.id_institucion },
        },
        tipoCarrito: { id_tipo_carrito: infoEquipo.id_tipo_carrito },
      },
      status: { id_status: infoEquipo.id_status },
    });
  }
}
